
.ltx_ulem_uline     { border-bottom:1px solid  black; }
.ltx_ulem_uuline    { border-bottom:3px double black; }
.ltx_ulem_dashuline { border-bottom:1px dashed black; }
.ltx_ulem_dotuline  { border-bottom:1px dotted black; }
.ltx_ulem_uwave     { border-bottom:3px groove black; } /*!?!!?!*/
.ltx_ulem_sout      { text-decoration:line-through; }
.ltx_ulem_xout      { text-decoration:line-through; } /*Should be multiple slashes! How??*/
